#!/usr/bin/python2.7
# -*- coding: utf-8 -*-
#
# Wait 300 seconds for the OX server to be responsive.
#
# Copyright (C) 2017 Univention GmbH
#
# http://www.univention.de/
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

import sys
import time
import logging
import subprocess
from univention.config_registry import ConfigRegistry
from univention.ox.listener_tools import ox_list_user, get_credentials as _read_credentials, ListenerTools_RuntimeError


_cred_cache = dict()
CSV_USER_ROW_USERNAME = 'Name'


def wait_for_ox_server(timeout=300, raise_exception=False, logfile=sys.stdout, context_id=None):  # type: (Optional[int], Optional[bool], Optional[file], Optional[str]) -> int
	"""
	Wait specified amount of time until the OX server is responsive.

	Returns (int) 0 if found running before timeout. If timeout is reached and
	the OX server is not running returns (int) >=1 (exit code of ox_call).

	If raise_exception=True instead of returning (int) 1 a
	ListenerTools_RuntimeError will be raised. The exception will have a member
	'stderr' with the error message of the last connection try.

	To catch the text output that'd go to stdout, use a file-like object:
		import StringIO
		out = StringIO.StringIO()
		wait_for_ox_server(logfile=out)
		out.seek(0)
		server_output = out.read()
		out.close()

	Return codes:
		  1 - Look-up failed. Service "OXUser_V2" is not available.
		101 - Authentication failed
		102 - Context does not exist
		105 - java.net.ConnectException: Verbindungsaufbau abgelehnt
		106 - SRV-0001 Categories=TRY_AGAIN Message=
			'The required service com.openexchange.threadpool.ThreadPoolService
			is temporary not available. Please try again later.'
		107 - no such object in table
		108 - com.openexchange.admin.rmi.exceptions.NoSuchObjectException:
			com.openexchange.admin.rmi.exceptions.NoSuchUserException: No such
			user .. in context ..
		112 - com.openexchange.admin.rmi.exceptions.NoSuchObjectException:
			com.openexchange.admin.rmi.exceptions.NoSuchGroupException: No such
			group .. in context ..
		501 - The server is running, but the contexts oxadmin user was not found.

	:param timeout: int: max seconds to wait for OX server, default=300
	:param raise_exception: bool: if True and OX server is not running
		raise ListenerTools_RuntimeError else return (int) 1, default=False
	:param logfile: Python file object: write output to this,
		default=sys.stdout
	:param context_id: str: context ID in which to look for oxadmin user,
		default=(UCRV ox/context/id | 10)
	:return: int: 0 if OX server is running or >=1 of not | raise
		ListenerTools_RuntimeError
	"""
	logger = logging.getLogger('wait_for_ox_server')  # black hole logger to suppress output from ox_list_user
	logger.handlers = [logging.NullHandler()]
	logger.setLevel(logging.ERROR)

	if not context_id:
		ucr = ConfigRegistry()
		ucr.load()
		context_id = ucr.get('ox/context/id', '10')

	try:
		admuser, admpass = _get_credentials(context_id)
	except ListenerTools_RuntimeError as exc:
		logfile.write('Falling back to allpluginsloaded. Error was: {}\n'.format(exc))
		admuser = admpass = None

	rows = list()
	retries = timeout
	while True:
		try:
			if admuser and admpass:
				rows = ox_list_user(admuser, admpass, context_id, searchpattern=admuser, logger=logger)
			else:
				run_allpluginsloaded()

			# need a line break before continuing
			if retries != timeout:
				logfile.write('\n')

			if admuser and admpass:
				for row in rows:
					if row[CSV_USER_ROW_USERNAME] == admuser:
						# while the server is shutting down, we can still use the HTTP-API
						# but allpluginsloaded finds missing (already unloaded) components
						run_allpluginsloaded()
						return 0
				if raise_exception:
					msg = 'The server is running, but the user "{}" was not found in context {}.'.format(
						admuser, context_id)
					raise ListenerTools_RuntimeError(msg, stderr=msg)
				else:
					return 501
			else:
				return 0
		except ListenerTools_RuntimeError as exc:
			if exc.exit_code in [1, 105, 106, 107] and retries > 0:
				if retries == timeout:
					logfile.write('Waiting {}s for OX server'.format(timeout))
				elif retries != timeout and retries % 10 == 0:
					logfile.write(str(retries))
				else:
					logfile.write('.')
				logfile.flush()
				time.sleep(1)
				retries -= 1
			else:
				logfile.write('{}\n'.format('0' if retries == 0 else ''))
				logfile.write(exc.stderr)
				if raise_exception:
					raise
				else:
					return exc.exit_code


def run_allpluginsloaded():  # type: () -> Tuple[int, str, str]
	"""
	Run /opt/open-xchange/sbin/allpluginsloaded.

	Raises ListenerTools_RuntimeError on error.

	:return: tuple(int, str, str): (exit code, stdout output, stderr output)
	"""
	cmd_proc = subprocess.Popen(
		['/opt/open-xchange/sbin/allpluginsloaded'],
		stdout=subprocess.PIPE,
		stderr=subprocess.PIPE,
		shell=False)
	cmd_out, cmd_err = cmd_proc.communicate()
	cmd_exit = cmd_proc.wait()
	if cmd_exit:
		raise ListenerTools_RuntimeError(
			'"allpluginsloaded" returned  {}: {}'.format(cmd_exit, cmd_out),
			cmd_exit,
			stderr=cmd_err)
	return cmd_exit, cmd_out.strip(), cmd_err.strip()


def _get_credentials(context_id):  # type: (str) -> Tuple[str, str]
	if not context_id in _cred_cache:
		if context_id == '10':
			admuser = 'oxadmin'
		else:
			admuser = 'oxadmin-context{}'.format(context_id)
		admpass = _read_credentials('/etc/ox-secrets/context{}.secret'.format(context_id))
		_cred_cache[context_id] = (admuser, admpass)
	return _cred_cache[context_id]


if __name__ == '__main__':
	sys.exit(wait_for_ox_server())
