#!/usr/bin/python2.7
# -*- coding: utf-8 -*-
#
#  OX for UCS
#  listener module: handling shared folders
#
# Copyright (C) 2017 Univention GmbH <http://www.univention.de/>
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

__package__ = ''  # workaround for PEP 366
import random
import string
from ldap.filter import filter_format
import listener
import univention.uldap
from univention.ox.oxlogger import OXLogger
from univention.ox.oxlistener import OxListener
from univention.ox.listener_tools import AsRoot, ListenerQueue, ListenerTools_RuntimeError, get_credentials_for_context, listener_is_deactivated, ox_call
try:
	from typing import Dict, List
except ImportError:
	pass


def folder_listener_is_activated():  # type: () -> bool
	return listener.configRegistry.is_true('mail/dovecot') and listener.configRegistry.is_true('ox/listener/folder/enabled')


name = 'ox-folder'
description = 'Open-Xchange support for shared folders'
filter = '(|(objectClass=univentionMailSharedFolder)(objectClass=oxSharedFolder))'
attributes = ['cn', 'mailPrimaryAddress']
modrdn = "1"

RENAME_TMP_PATH = '/var/spool/univention-directory-listener/ox_folder_modrdn.json'

logger = None

if listener_is_deactivated(name) or not folder_listener_is_activated():
	OXLogger(name).info('Listener module has been deactivated.')


def folder_context_id():  # type: () -> str
	return listener.configRegistry.get('ox/listener/folder/context', '') or listener.configRegistry.get('ox/context/id', '10')


def make_password():  # type: () -> str
	pw = [random.choice(string.ascii_letters + string.digits) for _x in range(30)]
	return ''.join(pw)


def get_auth_args():  # type: () -> List[str]
	contextid = folder_context_id()
	admuser, admpass = get_credentials_for_context(contextid)
	return [
		'--adminuser={}'.format(admuser),
		'--adminpass={}'.format(admpass),
		'--contextid={}'.format(contextid),
	]


def get_imap_server(domain):  # type: (str) -> str
	if listener.configRegistry.get('ox/role') != 'oxae':
		ldapbase = listener.configRegistry['ldap/base']
		with AsRoot():
			try:
				ldapCon = univention.uldap.getMachineConnection(ldap_master=False)
			except IOError as ex:
				raise ListenerTools_RuntimeError(ex)
		filter_s = filter_format('(&(cn=%s)(objectClass=oxMailDomainServerSettings))', (domain,))
		res = ldapCon.search(base='cn=domain,cn=mail,%s' % ldapbase, filter=filter_s, attr=['oxImapServer'])
		if res:
			return res[0][1].values()[0][0]
	return 'localhost'


def get_username_from_cn(cn):  # type: (str) -> str
	return 'sharedfolder_{}'.format(cn)


def get_ox_user_args(cn, mpa):  # type: (str, str) -> List[str]
	return [
		'--username={}'.format(get_username_from_cn(cn)),
		'--password={}'.format(make_password()),
		'--imaplogin={}'.format(mpa),
		'--givenname=Shared Folder',
		'--surname={}'.format(cn),
		'--email={}'.format(mpa),
		'--email1={}'.format(mpa),
		'--defaultsenderaddress={}'.format(mpa),
		'--aliases={}'.format(mpa),
		'--displayname=Shared Folder {}'.format(cn),
		'--mailenabled=true',
		'--access-combination-name=oxseforucs_webmail',
		'--access-usm=off'
		'--access-syncml=off',
		'--access-edit-resource=off',
		'--imapserver={}'.format(get_imap_server(mpa.rsplit('@')[-1])),
		'--smtpserver=localhost',
	]


def create_ox_user(cn, mpa):  # type: (str, str) -> None
	cmd = ['createuser']
	cmd += get_auth_args()
	cmd += get_ox_user_args(cn, mpa)
	ox_call(cmd, logger=logger)


def modify_ox_user(old_cn, old_mpa, new_cn, new_mpa):  # type: (str, str, str, str) -> None
	# Name (cn) change is not supported, as the UDM module forbids it.
	cmd = ['changeuser']
	cmd += get_auth_args()
	cmd += get_ox_user_args(new_cn, new_mpa)
	ox_call(cmd, logger=logger)


def remove_ox_user(old_cn, old_mpa):  # type: (str, str) -> None
	cmd = [
		'deleteuser',
		'--username={}'.format(get_username_from_cn(old_cn)),
	]
	cmd += get_auth_args()
	ox_call(cmd, 'com.openexchange.admin.rmi.exceptions.NoSuchObjectException', logger=logger)


@ListenerQueue(name, stop_on_error=False)
def handler(dn, new, old, command):
	# type: (str, Dict[str, List[str]], Dict[str, List[str]], str) -> None
	global logger

	if listener_is_deactivated(name) or not folder_listener_is_activated():
		return

	logger = OXLogger(name)

	if command == 'r':
		OxListener.modrdn_save_old(old, RENAME_TMP_PATH)
		logger.debug('command={!r} dn={!r}'.format(command, dn))
		return
	elif command == 'a':
		old = OxListener.modrdn_load_old(old, RENAME_TMP_PATH)

	new_cn = new.get('cn', [''])[0]
	old_cn = old.get('cn', [''])[0]
	new_mpa = new.get('mailPrimaryAddress', [None])[0]
	old_mpa = old.get('mailPrimaryAddress', [None])[0]

	if not new_mpa and not old_mpa:
		# nothing to do
		return

	logger.qualifier = new_cn or old_cn
	logger.debug('command={!r} dn={!r} old[mailPrimaryAddress]={!r} new[mailPrimaryAddress]={!r} old[cn]={!r} new[cn]={!r}'.format(
		command, dn, old_mpa, new_mpa, old_cn, new_cn))

	if new_mpa and not old_mpa:
		create_ox_user(new_cn, new_mpa)
		return

	if new_mpa and old_mpa:
		# Hint: Name (cn) change is not supported, as the UDM module forbids it.
		modify_ox_user(old_cn, old_mpa, new_cn, new_mpa)
		return

	if not new_mpa and old_mpa:
		remove_ox_user(old_cn, old_mpa)
		return


if __name__ == '__main__':
	handler.process_pending()
