#!/usr/bin/python2.7
# -*- coding: utf-8 -*-
#
#  OX for UCS
#  listener module: handling groups
#
# Copyright (C) 2004-2017 Univention GmbH <http://www.univention.de/>
# and
#     iKu Systems & Services GmbH & Co. KG (http://www.iku-systems.de/)
#
# All rights reserved.
#
# The source code of this program is made available
# under the terms of the GNU Affero General Public License version 3
# (GNU AGPL V3) as published by the Free Software Foundation.
#
# Binary versions of this program provided by Univention to you as
# well as other copyrighted, protected or trademarked materials like
# Logos, graphics, fonts, specific documentations and configurations,
# cryptographic keys etc. are subject to a license agreement between
# you and Univention and not subject to the GNU AGPL V3.
#
# In the case you use this program under the terms of the GNU AGPL V3,
# the program is provided in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public
# License with the Debian GNU/Linux or Univention distribution in file
# /usr/share/common-licenses/AGPL-3; if not, see
# <http://www.gnu.org/licenses/>.

__package__ = ''  # workaround for PEP 366
import os
from ldap import explode_dn
import listener
from univention.ox.oxlogger import OXLogger
from univention.ox.oxlistener import OxListener
from univention.ox.listener_tools import ListenerQueue, OXUserIdCache, OXGroupCache, ox_call, listener_is_deactivated, Exclusive, ListenerTools_RuntimeError

name = 'ox-groups'
description = 'Group settings for Open-Xchange'
filter = '(objectClass=posixGroup)'
attributes = []  # type: List[str]
modrdn = "1"


RENAME_TMP_PATH = '/var/spool/univention-directory-listener/ox_groups_modrdn.json'
# language
os.environ["LANG"] = listener.configRegistry.get("locale/default", "").split(':')[0]

logger = OXLogger(name)
_ox_group_cache = OXGroupCache(name)

try:
	with Exclusive():
		if __name__ != '__main__':   # do not rebuild cache if called by cron job
			# '__main__' ==> called by cron job
			# '/usr/lib/univention-directory-listener/system/ox-groups.py' ==> called by listener
			if listener.configRegistry.is_true('ox/listener/group-cache/autoinvalidate', True):
				_ox_group_cache.rebuild_cache(raise_exceptions=False)
		_ox_group_cache.rebuild_cache_if_required(raise_exceptions=False)
except ListenerTools_RuntimeError:
	logger.warn('could not get lock to auto-invalidate group cache - keeping old group cache')



class OXGroup(object):
	def __init__(self, groupname, context_ids):  # type: (str, List[str]) -> None
		self.groupname = groupname
		self.context_ids = context_ids
		self.displayname = dict()  # type: Dict[str, str]
		self.members = dict()   # type: Dict[str, List[str]]
		self.groupid = dict()   # type: Dict[str, str]

		for context_id in self.context_ids:
			for grp in _ox_group_cache.query_one_context(context_id, groupname=groupname):
				self.displayname[context_id] = grp.displayname
				self.members[context_id] = grp.member_list
				self.groupid[context_id] = grp.group_id

	def __repr__(self):  # type: () -> str
		return 'OXGroup(groupname={!r}, context_ids={!r}) displayname={!r} groupid={!r} members={!r}'.format(
			self.groupname, self.context_ids, self.displayname, self.groupid, self.members)


class OXGroupListener(OxListener):
	def __init__(self, name, old, new, command):  # type: (str, Dict[str, List[str]], Dict[str, List[str]], str) -> None
		super(OXGroupListener, self).__init__(name, old, new, command)

		self.old_context_ids = list()  # type: List[str]
		self.new_context_ids = list()  # type: List[str]
		self.groupname = self.get_groupname()
		self.logger.qualifier = self.groupname
		self.ox_group = None  # type: Union[None, OXGroup]
		self.new_members = dict()  # type: Dict[str, Dict[str, str]]
		self.old_members = dict()  # type: Dict[str, Dict[str, str]]
		self._ox_user_id_cache = OXUserIdCache(name)
		# check if OX user id cache has to be rebuild
		self._ox_user_id_cache.rebuild_cache_if_required()
		_ox_group_cache.rebuild_cache_if_required()


	def create_group(self, context_id):  # type: (str) -> None
		self.logger.process('Creating group "{}" in context {}...'.format(self.groupname, context_id))

		if not self.new_members:
			self.new_members = self.get_ox_members(self.new)
			self.logger.debug('OX users NEW by context: {}'.format(self.new_members))

		admuser, admpass = self.get_credentials(context_id)
		command = [
			'creategroup',
			'--displayname={}'.format(self.description),
		]
		if self.new_members.get(context_id):
			command.append('--addmembers={}'.format(','.join(self.new_members[context_id].values())))

		command.extend([
			'--adminuser={}'.format(admuser),
			'--adminpass={}'.format(admpass),
			'--contextid={}'.format(context_id),
			'--name={}'.format(self.groupname),
		])
		ox_call(command, logger=self.logger)
		_ox_group_cache.add_from_ox(context_id, self.groupname, admuser, admpass)

	def get_groupname(self):  # type: () -> str
		return self.new_groupname or self.old_groupname

	@property
	def old_groupname(self):  # type: () -> str
		return self.old.get('cn', [''])[0]

	@property
	def new_groupname(self):  # type: () -> str
		return self.new.get('cn', [''])[0]

	@property
	def description(self):  # type: () -> str
		res = self.new.get('description', [self.groupname])[0]
		return res[:128]  # OX length limit (Bug #35820)

	def get_ox_members(self, old_or_new):  # type: (Dict[str, List[str]]) -> Dict[str, Dict[str, str]]
		"""
		Get all usernames and user-ids from all contexts for this groups members.

		:param old_or_new: dict: listener new/old
		:return: dict: {context id: {username: ox_user_id, ..}, ..}
		Example: {'280': {'test19': '31', 'user28': '3'}, '12': {'user32': '39'}, '10': {'test30': '179', 'user31': '162'}}
		"""
		ox_users = dict()  # type: Dict[str, Dict[str, str]]
		try:
			member_dns = old_or_new['uniqueMember']
		except (KeyError, TypeError):
			return ox_users

		# collect context IDs and usernames of members
		for dn in member_dns:
			if not dn.startswith('uid='):
				continue
			username = explode_dn(dn, 1)[0]
			cache_entries = self._ox_user_id_cache.query(username=username)

			for context, username_and_id in cache_entries.items():
				for ox_id in username_and_id.values():
					ox_users.setdefault(context, dict())[username] = ox_id
		return ox_users

	def modify_group(self, context_id):  # type: (str) -> None
		self.logger.process('Modifying group "{}" in context {}...'.format(self.groupname, context_id))

		if not self.new_members:
			self.new_members = self.get_ox_members(self.new)
			self.logger.debug('OX group members NEW by context: {}'.format(self.new_members))
		if not self.old_members:
			self.old_members = self.get_ox_members(self.old)
			self.logger.debug('OX group members OLD by context: {}'.format(self.old_members))

		add_members = set(self.new_members.get(context_id, {}).values()) - set(self.old_members.get(context_id, {}).values())
		add_members.difference_update(set(self.ox_group.members.get(context_id, [])))  # don't try to add existing members
		rm_members = set(self.old_members.get(context_id, {}).values()) - set(self.new_members.get(context_id, {}).values())

		if all((
				not add_members,
				not rm_members,
				self.ox_group.displayname[context_id] == self.description,
				self.ox_group.groupname == self.groupname)):
			self.logger.process('No changes for OX group "{}" in context {}.'.format(self.groupname, context_id))
			return

		admuser, admpass = self.get_credentials(context_id)
		command = [
			'changegroup',
			'--displayname={}'.format(self.description),
		]
		if rm_members:
			command.append("--removemembers=%s" % ','.join(rm_members))
		if add_members:
			command.append("--addmembers=%s" % ','.join(add_members))
		command.extend([
			'--adminuser={}'.format(admuser),
			'--adminpass={}'.format(admpass),
			'--contextid={}'.format(context_id),
			'--name={}'.format(self.groupname),
			'--groupid={}'.format(self.ox_group.groupid[context_id]),
		])
		ox_call(command, logger=self.logger)
		# update ox group cache
		grp = None
		try:
			grp = _ox_group_cache.query_one_context(context_id, group_id=self.ox_group.groupid[context_id])[0]
		except IndexError:
			self.logger.error('Group has been renamed but was unable to find old group in group cache: group=%r context=%r' % (self.groupname, context_id))
		if grp is not None:
			if grp.groupname != self.groupname:
				self.logger.info('Group has been renamed - updating group cache: %r --> %r' % (grp.groupname, self.groupname))
				_ox_group_cache.rename_group_by_id(context_id, self.groupname, self.ox_group.groupid[context_id])
			_ox_group_cache.add_members(context_id, self.groupname, add_members)
			_ox_group_cache.remove_members(context_id, self.groupname, rm_members)

	@property
	def new_is_ox_group(self):  # type: () -> bool
		return self.new.get('isOxGroup', [''])[0].lower() == 'ok'

	@property
	def old_is_ox_group(self):  # type: () -> bool
		return self.old.get('isOxGroup', [''])[0].lower() == 'ok'

	def run(self):  # type: () -> None
		# ignore groups with name "users" (Bug #35821)
		if self.groupname.lower() == "users":
			self.logger.process('Ignoring group "{}".'.format(self.groupname))
			return

		if self.old:
			self.old_context_ids = self.get_context_ids_from_group(self.old)
			self.logger.process('Context IDs of group "{}" in old: {}'.format(self.old_groupname, self.old_context_ids))
		else:
			self.old_context_ids = list()
			self.logger.process('No old.')
		if self.new:
			self.new_context_ids = self.get_context_ids_from_group(self.new)
			self.logger.process('Context IDs of group "{}" in new: {}'.format(self.new_groupname, self.new_context_ids))
		else:
			self.new_context_ids = list()
			self.logger.process('No new.')

		if self.new_is_ox_group:
			# remove group from missing contexts
			for context_id in set(self.old_context_ids) - set(self.new_context_ids):
				self.remove_group(self.old_groupname or self.new_groupname, context_id)

			if self.old_is_ox_group and self.old_groupname != self.new_groupname:
				# rename
				self.ox_group = OXGroup(self.old_groupname, self.old_context_ids)
				self.logger.process('Renaming {} to "{}"...'.format(self.ox_group, self.groupname))
				for context_id in self.old_context_ids:
					if context_id in self.ox_group.members and context_id in self.new_context_ids:
						self.modify_group(context_id)

			self.ox_group = OXGroup(self.groupname, self.new_context_ids)
			self.logger.debug(self.ox_group)

			for context_id in self.new_context_ids:
				if context_id in self.ox_group.members:
					# group exists in context
					self.modify_group(context_id)
				else:
					self.create_group(context_id)
			return

		# remove group from all contexts
		if self.old_is_ox_group and not self.new_is_ox_group:
			for context_id in self.old_context_ids:
				self.remove_group(self.groupname, context_id)
			return


@ListenerQueue(name, stop_on_error=False)
def handler(dn, new, old, command):
	# type: (str, Dict[str, List[str]], Dict[str, List[str]], str) -> None
	if listener_is_deactivated(name):
		return
	if command == 'r':
		OxListener.modrdn_save_old(old, RENAME_TMP_PATH)
		logger.debug('command={!r} dn={!r}'.format(command, dn))
		return
	elif command == 'a':
		old = OxListener.modrdn_load_old(old, RENAME_TMP_PATH)
	logger.debug('command={!r} dn={!r}'.format(command, dn))

	ox_group_listener = OXGroupListener(name, old, new, command)
	ox_group_listener.run()


if __name__ == '__main__':
	handler.process_pending()
